package org.xbib.interlibrary.sru;

import java.io.IOException;
import java.io.InputStream;
import java.io.UncheckedIOException;
import java.nio.charset.StandardCharsets;
import java.security.SecureRandom;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Random;
import java.util.function.Consumer;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.common.MultiMap;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.datastructures.tiny.TinyMultiMap;
import org.xbib.interlibrary.action.avail.AbstractAvailService;
import org.xbib.interlibrary.action.avail.AvailRequest;
import org.xbib.interlibrary.action.avail.AvailResponseBuilder;
import org.xbib.interlibrary.action.avail.AvailServiceResponse;
import org.xbib.interlibrary.action.avail.AvailServiceResponseBuilder;
import org.xbib.interlibrary.api.library.Library;
import org.xbib.interlibrary.api.library.LibraryService;
import org.xbib.interlibrary.api.library.DefaultLibraryService;
import org.xbib.interlibrary.api.action.Response;
import org.xbib.interlibrary.api.action.ServiceArguments;
import org.xbib.interlibrary.api.library.Material;
import org.xbib.interlibrary.api.library.Mode;
import org.xbib.marc.Marc;
import org.xbib.marc.MarcRecord;
import org.xbib.marc.MarcRecordListener;
import org.xbib.marc.MarcXchangeConstants;
import org.xbib.marc.xml.MarcContentHandler;
import org.xbib.settings.Settings;
import static org.xbib.interlibrary.api.library.Material.ONLINE_RESOURCE;

public abstract class AbstractSRUAvailService extends AbstractAvailService {

    private static final Logger logger = Logger.getLogger(AbstractSRUAvailService.class.getName());

    private static final Random RANDOM = new SecureRandom();

    protected final SRUClientHelper clientHelper;

    protected final MarcHelper marcHelper;

    protected final Settings settings;

    protected AbstractSRUAvailService(ServiceArguments arguments, SRUClientHelper clientHelper) {
        super(arguments);
        this.clientHelper = clientHelper;
        this.marcHelper = MarcHelper.INSTANCE;
        this.settings = arguments.getSettings().getAsSettings("sru." + arguments.getName());
    }

    @Override
    public void execute(AvailRequest availRequest, AvailResponseBuilder availResponse) {
        AvailServiceResponseBuilder availServiceResponse = AvailServiceResponse.builder();
        if (!availRequest.isValidated()) {
            logger.log(Level.FINE, "request not validated, returning empty response");
            availServiceResponse.setStatus(getName(), Response.Status.EMPTY_RESPONSE);
            return;
        }
        availServiceResponse.setRequest(availRequest);
        availServiceResponse.setService(getName());
        try {
            String source = availRequest.getBibliographicDescription().getSource();
            if (supports(source) == null) {
                logger.log(Level.WARNING, this.getClass().getName() + " is unable to support " + source);
                availServiceResponse.setStatus("global", Response.Status.UNDEFINED);
                return;
            }
            MultiMap<String, LibraryService> multiMap = new TinyMultiMap<>();
            String query = getResolveCQL(availRequest);
            if (query == null) {
                return;
            }
            logger.log(Level.FINE, () -> getName() + ": executing avail request = " + availRequest + " cql = '" + query + "'");
            clientHelper.searchCQL(query,
                    parameterCompletion(),
                    httpResponse -> {
                        try {
                            process(availRequest, availServiceResponse, httpResponse.getBodyAsStream(), multiMap);
                        } catch (Exception e) {
                            logger.log(Level.SEVERE, e.getMessage(), e);
                            availServiceResponse.setStatus(getName(), Response.Status.ERROR);
                            availServiceResponse.setMessage(e.getMessage());
                            availServiceResponse.setThrowable(e);
                            throw new IllegalStateException(e);
                        }
                    },
                    exception -> {
                        logger.log(Level.SEVERE, exception.getMessage(), exception);
                        availServiceResponse.setStatus(getName(), Response.Status.ERROR);
                        availServiceResponse.setMessage(exception.getMessage());
                        availServiceResponse.setThrowable(exception);
                        throw new IllegalStateException(exception);
                    },
                    req -> {
                        availServiceResponse.setStatus(getName(), Response.Status.TIMEOUT);
                        logger.log(Level.WARNING, getName() + " timeout");
                    });
            if (multiMap.isEmpty() && availServiceResponse.getStatus() == Response.Status.UNDEFINED) {
                availServiceResponse.setStatus(getName(), Response.Status.EMPTY_RESPONSE);
            } else {
                toResult(availRequest, availServiceResponse, multiMap);
                availServiceResponse.setStatus(getName(), Response.Status.OK);
            }
        } catch (Throwable e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            availServiceResponse.setStatus(getName(), Response.Status.ERROR);
            availServiceResponse.setMessage(e.getMessage());
            availServiceResponse.setThrowable(e);
            throw new IllegalStateException(e);
        } finally {
            availResponse.add(availServiceResponse);
        }
    }

    @Override
    public void close() throws IOException {
        clientHelper.close();
    }

    @Override
    public String getURL(String id) {
        return arguments.getServiceSettings("url", "");
    }

    @Override
    public String getLabel() {
        return arguments.getServiceSettings("label", getName().toUpperCase(Locale.ROOT));
    }

    @Override
    public String getSortKey() {
        return arguments.getServiceSettings("sort", "00");
    }

    protected abstract Consumer<TinyMap.Builder<String, Object>> parameterCompletion();

    protected String getResolveCQL(AvailRequest availRequest) {
        String source = availRequest.getBibliographicDescription().getSource();
        if (source == null) {
            logger.log(Level.FINE, getName() + ": source is null, rejected");
            return null;
        }
        String sourceId = availRequest.getBibliographicDescription().getSourceId();
        if (sourceId == null) {
            logger.log(Level.FINE, getName() + ": source ID is null, rejected ");
            return null;
        }
        source = supports(source);
        if (source == null) {
            logger.log(Level.FINE, getName() + ": source is not supported: " + source);
            return null;
        }
        return sourceIdCQL(normalizeSourceId(sourceId));
    }

    protected abstract String supports(String source);

    protected String normalizeSourceId(String sourceId) {
        return sourceId.toLowerCase(Locale.ROOT).trim().replace("-", "");
    }

    protected String sourceIdCQL(String sourceId) {
        return "bib.identifierLocalNumber = " + sourceId;
    }

    protected List<LibraryService> extractGermanInterlibraryServices(AvailRequest availRequest,
                                                                     MarcRecord marcRecord) {
        final List<LibraryService> list = new LinkedList<>();
        String identifier = identifier(marcRecord);
        List<Material> materials = marcHelper.extractMaterial(marcRecord);
        marcRecord.getAll("924").forEach(marcField -> {
            DefaultLibraryService.Builder serviceBuilder = DefaultLibraryService.builder();
            serviceBuilder.setId(Integer.toString(random()));
            DefaultLibraryService.DefaultLocation location = new DefaultLibraryService.DefaultLocation();
            StringBuilder url = new StringBuilder();
            StringBuilder label = new StringBuilder();
            // Default value for "mode". We put it here as long as GVI data does not give the real mode.
            serviceBuilder.setMode(Mode.LOAN_COPY);
            marcField.getSubfields().forEach(subfield -> {
                switch (subfield.getId()) {
                    case "9":
                        // sigel, skip
                        break;
                    case "b":
                        // ISIL
                        String isil = subfield.getValue();
                        Library library = configuration.getLibrary(isil);
                        if (library != null) {
                            serviceBuilder.setISIL(library.getISIL());
                            serviceBuilder.setName(library.getName());
                            serviceBuilder.setDomainPriority(domainPriority(availRequest, library));
                            serviceBuilder.setLibraryPriority(libraryPriority(availRequest, library));
                            serviceBuilder.setServicePriority(materials.contains(ONLINE_RESOURCE) ? 1 : 2);
                            serviceBuilder.setSublibraryPriority(subPriority(library));
                            serviceBuilder.setMaterials(materials);
                            if (library.getDomainName() != null) {
                                serviceBuilder.setSource(library.getDomainName());
                            }
                            serviceBuilder.setSourceId(identifier);
                        } else {
                            logger.log(Level.WARNING, "ISIL not found in configuration: " + isil);
                        }
                        break;
                    case "c":
                        // domain
                        // skip, use the configured one derived from our library setup
                        break;
                    case "d":
                        // loan indicator
                        switch (subfield.getValue()) {
                            case "L", "a" -> serviceBuilder.setMode(Mode.LOAN);
                            case "C", "b" -> serviceBuilder.setMode(Mode.COPY);
                            case "LC", "c" -> serviceBuilder.setMode(Mode.LOAN_COPY);
                            case "N", "d" -> serviceBuilder.setMode(Mode.NONE);
                            case "e" -> serviceBuilder.setMode(Mode.ELECTRONIC_COPY);
                            default -> {
                            }
                        }
                        break;
                    case "e":
                        // loan restrictions
                        // n, p
                        //switch (subfield.getValue()) {
                        //    case "n" -> map.put("restrictions", "none");
                        //    case "p" -> map.put("restrictions", "postal");
                        //}
                        break;
                    case "g":
                        // call number
                        location.setCallnumber(subfield.getValue());
                        break;
                    case "h":
                        location.setCollection(subfield.getValue());
                        break;
                    case "k":
                        // URL
                        if (!url.isEmpty()) {
                            location.addLink(url.toString(), label.toString());
                            url.setLength(0);
                            label.setLength(0);
                        }
                        url.append(subfield.getValue());
                        break;
                    case "l":
                        // URL label
                        label.append(subfield.getValue());
                        break;
                    case "z":
                        // textual holdings
                        location.setTextualHoldings(trimBrackets(subfield.getValue()));
                        break;
                    default:
                        break;
                }
            });
            if (!url.isEmpty()) {
                location.addLink(url.toString(), label.toString());
                url.setLength(0);
                label.setLength(0);
            }
            serviceBuilder.addLocation(location);
            list.add(serviceBuilder.build());
        });
        return list;
    }

    protected String identifier(MarcRecord marcRecord) {
        return trim(marcRecord.getFirst("001").getValue());
    }

    private void process(AvailRequest request,
                           AvailServiceResponseBuilder response,
                           InputStream inputStream,
                           MultiMap<String, LibraryService> multiMap) {
        MarcContentHandler marcContentHandler = new MarcContentHandler();
        marcContentHandler.setMarcRecordListener(getMarcRecordListener(request, response, multiMap));
        try {
            Marc.builder()
                    .setInputStream(inputStream)
                    .setCharset(StandardCharsets.UTF_8)
                    .setContentHandler(new MarcContentHandler()
                            .setFormat("MarcXML")
                            .setType(MarcXchangeConstants.BIBLIOGRAPHIC_TYPE)
                            .addNamespace("http://www.loc.gov/MARC21/slim")
                            .setMarcListener(MarcXchangeConstants.BIBLIOGRAPHIC_TYPE, marcContentHandler)
                            .setMarcListener(MarcXchangeConstants.HOLDINGS_TYPE, marcContentHandler))
                    .build()
                    .xmlReader()
                    .parse();
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    protected abstract MarcRecordListener getMarcRecordListener(AvailRequest availRequest,
                                                                AvailServiceResponseBuilder availResponse,
                                                                MultiMap<String, LibraryService> multiMap);

    protected static Integer random() {
        return RANDOM.nextInt();
    }

    private static String trimBrackets(String string) {
        return string.startsWith("<") && string.endsWith(">") ? string.substring(1, string.length() - 1) : string;
    }

    private static String trim(String string) {
        return string != null ? string.trim() : null;
    }
}
