package org.xbib.interlibrary.sru;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.function.Consumer;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.interlibrary.action.resolve.AbstractResolveService;
import org.xbib.interlibrary.action.resolve.ResolveRequest;
import org.xbib.interlibrary.action.resolve.ResolveResponseBuilder;
import org.xbib.interlibrary.action.resolve.ResolveServiceResponse;
import org.xbib.interlibrary.action.resolve.ResolveServiceResponseBuilder;
import org.xbib.interlibrary.api.BibliographicDescription;
import org.xbib.interlibrary.api.action.Response;
import org.xbib.interlibrary.api.action.ServiceArguments;
import org.xbib.interlibrary.common.DefaultBibliographicDescription;
import org.xbib.settings.Settings;
import org.xbib.standardnumber.ISBN;
import org.xbib.standardnumber.ISSN;
import org.xbib.standardnumber.StandardNumber;
import org.xbib.standardnumber.ZDB;

public abstract class AbstractSRUResolveService extends AbstractResolveService {

    private static final Logger logger = Logger.getLogger(AbstractSRUResolveService.class.getName());

    protected final SRUClientHelper clientHelper;

    protected final MarcHelper marcHelper;

    protected final Settings settings;

    protected AbstractSRUResolveService(ServiceArguments arguments, SRUClientHelper clientHelper) {
        super(arguments);
        this.clientHelper = clientHelper;
        this.marcHelper = MarcHelper.INSTANCE;
        this.settings = arguments.getSettings().getAsSettings("sru." + arguments.getName());
    }

    @Override
    public void execute(ResolveRequest request, ResolveResponseBuilder response) {
        ResolveServiceResponseBuilder serviceResponseBuilder = ResolveServiceResponse.builder();
        serviceResponseBuilder.setService(getName());
        try {
            BibliographicDescription bibliographicDescription = request.getBibliographicDescription();
            String source = bibliographicDescription.getSource();
            String sourceId = normalizeSourceId(bibliographicDescription.getSourceId());
            // default value for formatted source ID
            String formattedSourceId = bibliographicDescription.getFormattedSourceId() != null ?
                    bibliographicDescription.getFormattedSourceId() : bibliographicDescription.getSourceId();
            // we have to clean up the source ID to our reduced format in lower case and without hyphens
            // because it may come from external 'dirty' sources
            if ("zdb".equalsIgnoreCase(source)) {
                ZDB zdb = new ZDB();
                try {
                    boolean fixChecksum = bibliographicDescription.getChecksum();
                    if (fixChecksum) {
                        sourceId = sourceId + "0"; // add dummy checksum
                    }
                    zdb.set(sourceId).createChecksum(fixChecksum).normalize().verify();
                } catch (NumberFormatException e) {
                    logger.log(Level.WARNING, "invalid ZDB detected: " + sourceId);
                }
                sourceId = zdb.normalizedValue();
                formattedSourceId = zdb.format();
            }
            logger.log(Level.FINER, "original source id = " + bibliographicDescription.getSourceId() +
                    " normalized = " + sourceId +
                    " formatted = " + formattedSourceId +
                    " checksum = " + bibliographicDescription.getChecksum());
            DefaultBibliographicDescription.Builder builder = DefaultBibliographicDescription.builder()
                    .setSource(source)
                    .setSourceId(sourceId)
                    .setFormattedSourceId(formattedSourceId)
                    .setChecksum(bibliographicDescription.getChecksum())
                    .setIdentifiers(bibliographicDescription.getIdentifiers())
                    .setAuthor(bibliographicDescription.getAuthor())
                    .setTitle(bibliographicDescription.getTitle())
                    .setInTitle(bibliographicDescription.getInTitle())
                    .setPublisher(bibliographicDescription.getPublisher())
                    .setYear(bibliographicDescription.getYear())
                    .setVolume(bibliographicDescription.getVolume())
                    .setIssue(bibliographicDescription.getIssue())
                    .setPagination(bibliographicDescription.getPagination());
            String query = getResolveCQL(source, sourceId, formattedSourceId);
            if (query == null) {
                logger.log(Level.FINE, getName() + ": resolve query is null, rejected (source = " + source + ", sourceId = " + sourceId + ")");
                return;
            }
            logger.log(Level.FINE, getName() + ": executing resolve request = " + request.asMap() + " CQL query = " + query);
            clientHelper.searchCQL(query,
                    parameterCompletion(),
                    httpResponse -> {
                        try (InputStream inputStream = httpResponse.getBodyAsStream()) {
                            resolve(request, serviceResponseBuilder, inputStream, builder);
                            serviceResponseBuilder.setRequest(request);
                            serviceResponseBuilder.setBibliographicDescription(builder.build());
                        } catch (Exception e) {
                            logger.log(Level.SEVERE, e.getMessage(), e);
                            serviceResponseBuilder.setStatus(getName(), Response.Status.ERROR);
                            throw new IllegalStateException(e);
                        }
                    },
                    exception -> {
                        logger.log(Level.SEVERE, exception.getMessage(), exception);
                        serviceResponseBuilder.setStatus(getName(), Response.Status.ERROR);
                        throw new IllegalStateException(exception);
                    },
                    req -> {
                        serviceResponseBuilder.setStatus(getName(), Response.Status.TIMEOUT);
                        logger.log(Level.WARNING, "timeout");
                    });
        } catch (Throwable e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            serviceResponseBuilder.setStatus(getName(), Response.Status.ERROR);
            throw new IllegalStateException(e);
        } finally {
            response.add(serviceResponseBuilder);
        }
    }

    @Override
    public void close() throws IOException {
        clientHelper.close();
    }
    protected String normalizeSourceId(String sourceId) {
        return sourceId != null ? sourceId.toLowerCase(Locale.ROOT).trim().replace("-", "") : null;
    }

    protected abstract Consumer<TinyMap.Builder<String, Object>> parameterCompletion();

    protected String getResolveCQL(String source, String sourceId, String formattedSourceId) {
        String supportedSource = supports(source);
        if (supportedSource == null) {
            return null;
        }
        return sourceIdCQL(sourceId, formattedSourceId);
    }

    protected String sourceIdCQL(String sourceId, String formattedSourceId) {
        return "bib.identifierLocalNumber = " + sourceId;
    }

    protected abstract String supports(String source);

    protected abstract void resolve(ResolveRequest resolveRequest,
                                    ResolveServiceResponseBuilder serviceResponseBuilder,
                                    InputStream inputStream,
                                    DefaultBibliographicDescription.Builder builder);

    protected void addIdentifier(DefaultBibliographicDescription.Builder builder, String type, Object object) {
        StandardNumber standardNumber = switch (type) {
            case "ISBN" -> new ISBN();
            case "ISSN" -> new ISSN();
            default -> null;
        };
        for (Object value : toList(object)) {
            if (standardNumber != null && value != null) {
                try {
                    standardNumber.set(value.toString()).normalize().verify();
                    if (standardNumber.isValid() && standardNumber.normalizedValue() != null && standardNumber.format() != null) {
                        builder.addIdentifier(Map.of("type", type,
                                "value", standardNumber.normalizedValue(),
                                "formatted", standardNumber.format()));
                    }
                } catch (NumberFormatException e) {
                    // ignore wrong ISBNs/ISSNs
                }
            }
        }
    }

    @SuppressWarnings("unchecked")
    private static List<Object> toList(Object object) {
        if (object == null) {
            return List.of();
        }
        if (object instanceof List) {
            return (List<Object>) object;
        }
        return List.of(object);
    }
}
