package org.xbib.interlibrary.sru;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UncheckedIOException;
import java.nio.charset.StandardCharsets;
import java.util.concurrent.atomic.AtomicLong;
import java.util.function.Consumer;
import java.util.function.Supplier;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.interlibrary.action.search.AbstractSearchService;
import org.xbib.interlibrary.action.search.SearchRequest;
import org.xbib.interlibrary.action.search.SearchResponse;
import org.xbib.interlibrary.action.search.SearchResponseBuilder;
import org.xbib.interlibrary.action.search.SearchService;
import org.xbib.interlibrary.action.search.SearchSyntax;
import org.xbib.interlibrary.api.action.Request;
import org.xbib.interlibrary.api.action.Response;
import org.xbib.interlibrary.api.action.ServiceArguments;
import org.xbib.marc.Marc;
import org.xbib.marc.MarcRecord;
import org.xbib.marc.MarcRecordListener;
import org.xbib.marc.xml.MarcContentHandler;
import org.xbib.settings.Settings;
import static org.xbib.interlibrary.action.search.SearchSyntax.CQL;
import static org.xbib.interlibrary.action.search.SearchSyntax.PQF;

public abstract class AbstractSRUSearchService extends AbstractSearchService<SearchRequest, SearchResponse<SearchRequest>, SearchService> {

    private static final Logger logger = Logger.getLogger(AbstractSRUSearchService.class.getName());

    protected final SRUClientHelper clientHelper;

    protected final MarcHelper marcHelper;

    protected final Settings settings;

    protected final CQLHelper cqlHelper;

    protected AbstractSRUSearchService(ServiceArguments arguments, SRUClientHelper clientHelper) {
        super(arguments);
        this.clientHelper = clientHelper;
        this.marcHelper = MarcHelper.INSTANCE;
        this.settings = arguments.getSettings().getAsSettings("sru." + arguments.getName());
        this.cqlHelper = new CQLHelper(settings);
    }

    @Override
    public void execute(SearchRequest searchRequest, SearchResponseBuilder response) {
        final AtomicLong count = new AtomicLong(0);
        try {
            switch (searchRequest.getSearchSyntax()) {
                case PQF -> {
                    String cql = cqlHelper.fromPQFQuery(searchRequest);
                    logger.log(Level.FINE, getName() + ": cql query from pqf = " + cql);
                    SearchRequest request = cqlHelper.mapSearchRequest(searchRequest, cql);
                    String query = cqlHelper.fromCQLQuery(request);
                    logger.log(Level.FINE, getName() + ": cql query after mapped search= " + query);
                    clientHelper.searchCQL(query,
                            parameterCompletion(request),
                            httpResponse -> {
                                try {
                                    process(request, response, count, httpResponse.getStatus().code(), httpResponse.getBodyAsStream());
                                } catch (Throwable e) {
                                    response.setStatus(getName(), Response.Status.ERROR);
                                    response.setThrowable(e);
                                    logger.log(Level.WARNING, e.getMessage(), e);
                                    throw new UncheckedIOException(new IOException(e));
                                }
                            },
                            exception -> {
                                response.setStatus(getName(), Response.Status.ERROR);
                                response.setThrowable(exception);
                                logger.log(Level.SEVERE, exception.getMessage(), exception);
                                throw new UncheckedIOException(new IOException(exception));
                            },
                            req -> {
                                response.setStatus(getName(), Response.Status.TIMEOUT);
                                response.addMetadata(getName(), "total", 0L);
                                logger.log(Level.WARNING, getName() + ": timeout");
                            });
                }
                case CQL -> {
                    String cql = cqlHelper.fromCQLQuery(searchRequest);
                    logger.log(Level.FINE, getName() + ": cql query from cql = " + cql);
                    SearchRequest request = cqlHelper.mapSearchRequest(searchRequest, cql);
                    String query = cqlHelper.fromCQLQuery(request);
                    logger.log(Level.FINE, getName() + ": cql query after mapped search  = " + query);
                    clientHelper.searchCQL(query,
                            parameterCompletion(request),
                            httpResponse -> {
                                try {
                                    process(request, response, count, httpResponse.getStatus().code(), httpResponse.getBodyAsStream());
                                } catch (Throwable e) {
                                    response.setStatus(getName(), Response.Status.ERROR);
                                    response.setThrowable(e);
                                    logger.log(Level.WARNING, e.getMessage(), e);
                                    throw new UncheckedIOException(new IOException(e));
                                }
                            },
                            exception -> {
                                response.setStatus(getName(), Response.Status.ERROR);
                                response.setThrowable(exception);
                                logger.log(Level.SEVERE, exception.getMessage(), exception);
                                throw new UncheckedIOException(new IOException(exception));
                            },
                            req -> {
                                response.setStatus(getName(), Response.Status.TIMEOUT);
                                response.addMetadata(getName(), "total", 0L);
                                logger.log(Level.WARNING, getName() + ": timeout");
                            });
                }
                default -> {
                    throw new UnsupportedOperationException();
                }
            }
        } catch (Throwable e) {
            response.setStatus(getName(), Response.Status.ERROR);
            logger.log(Level.WARNING, e.getMessage(), e);
        }
    }

    @Override
    public void close() throws IOException {
        super.close();
        clientHelper.close();
    }

    @Override
    public boolean supports(SearchSyntax searchSyntax) {
        return CQL == searchSyntax || PQF == searchSyntax;
    }

    protected void process(SearchRequest searchRequest,
                           SearchResponseBuilder searchResponse,
                           AtomicLong count,
                           int statusCode,
                           InputStream bodyStream) throws IOException {
        long t0 = System.currentTimeMillis();
        if (statusCode != 200) {
            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
            bodyStream.transferTo(outputStream);
            logger.log(Level.WARNING, getName() + ": unexpected status " + statusCode +
                    " response = " + outputStream.toString(StandardCharsets.UTF_8));
            return;
        }
        MarcContentHandler marcContentHandler = new MarcContentHandler();
        marcContentHandler.setMarcRecordListener(getMarcRecordListener(searchRequest, searchResponse, count));
        SRUContentHandler sruContentHandler = getContentHandler(marcContentHandler);
        Marc.builder()
                .setInputStream(bodyStream)
                .setCharset(StandardCharsets.UTF_8)
                .setContentHandler(sruContentHandler)
                .build()
                .xmlReader()
                .parse();
        long total = sruContentHandler.getNumberOfRecords();
        searchResponse.addMetadata(getName(), "total", total);
        if (total > 0L) {
            searchResponse.addMetadata(getName(), "took", System.currentTimeMillis() - t0);
            searchResponse.addMetadata(getName(), "age", searchRequest.getAge());
            searchResponse.setStatus(getName(), Response.Status.OK);
        } else {
            searchResponse.setStatus(getName(), Response.Status.EMPTY_RESPONSE);
        }
    }

    protected Consumer<TinyMap.Builder<String, Object>> parameterCompletion(Request request) {
        return params -> {
            int offset = fixByOne() ? request.getOffset() + 1 : request.getOffset();
            params.put("startRecord", offset);
            params.put("maximumRecords", request.getSize());
            params.put("stylesheet", "");
        };
    }

    protected abstract MarcRecordListener getMarcRecordListener(SearchRequest searchRequest,
                                                                SearchResponseBuilder searchResponse,
                                                                AtomicLong count);
    protected abstract Supplier<String> getIdSupplier(MarcRecord marcRecord);

    protected abstract SRUContentHandler getContentHandler(MarcContentHandler marcContentHandler);
}
