package org.xbib.interlibrary.sru;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.cql.BooleanGroup;
import org.xbib.cql.CQLGenerator;
import org.xbib.cql.CQLParser;
import org.xbib.cql.Identifier;
import org.xbib.cql.Index;
import org.xbib.cql.Modifier;
import org.xbib.cql.ModifierList;
import org.xbib.cql.PrefixAssignment;
import org.xbib.cql.Query;
import org.xbib.cql.Relation;
import org.xbib.cql.ScopedClause;
import org.xbib.cql.SearchClause;
import org.xbib.cql.SimpleName;
import org.xbib.cql.SingleSpec;
import org.xbib.cql.SortSpec;
import org.xbib.cql.SortedQuery;
import org.xbib.cql.Term;
import org.xbib.cql.Visitor;
import org.xbib.interlibrary.action.search.SearchRequest;
import org.xbib.interlibrary.action.search.SearchSyntax;
import org.xbib.settings.Settings;

public class CQLHelper {

    private static final Logger logger = Logger.getLogger(CQLHelper.class.getName());

    private final TermTransformer termTransformer;

    public CQLHelper(Settings settings) {
        Map<String, String> keys = settings.getByPrefix("keys.").getAsMap();
        this.termTransformer = new TermTransformer(keys);
    }

    public SearchRequest mapSearchRequest(SearchRequest searchRequest,
                                          String query) {
        CQLParser cqlParser = new CQLParser(query);
        cqlParser.parse();
        CQLGenerator cqlGenerator = new CQLGenerator();
        cqlGenerator.setSubstitutor(termTransformer);
        cqlParser.getCQLQuery().accept(cqlGenerator);
        SearchRequest newSearchRequest = SearchRequest.builder()
                .setSearchSyntax(SearchSyntax.CQL)
                .setSearchOp(List.of(""))
                .setSearchKey(List.of(""))
                .setSearchValue(List.of(cqlGenerator.getResult()))
                .setOffset(searchRequest.getOffset())
                .setSize(searchRequest.getSize())
                .setSourceEnabled(searchRequest.isSourceEnabled())
                .setExtended(searchRequest.isExtended())
                .setDebug(searchRequest.isDebug())
                .setValidated(true)
                .build();
        logger.log(Level.FINE, "request offset = " + searchRequest.getOffset() +
                " cql query = " + query +
                " validated cql query = " + cqlGenerator.getResult() +
                " new search request = " + newSearchRequest);
        return newSearchRequest;
    }

    public String fromPQFQuery(SearchRequest request) {
        // abbreviated: single query parameter
        if (request.getSearchKey().size() == 1 &&
                request.getSearchKey().getFirst().isEmpty() &&
                request.getSearchValue().size() == 1) {
            return trim(request.getSearchValue().getFirst());
        }
        // PQF is just CQL! Map search key/op/value from form to a CQL string, skip attr mapping.
        LinkedList<String> termList = new LinkedList<>();
        for (int i = 0; i < request.getSearchValue().size(); i++) {
            String value = request.getSearchValue().get(i);
            value = trim(value);
            if (!value.isEmpty()) {
                String key = request.getSearchKey().get(i);
                String term = key + " " + getRelation() + " " + quote(value);
                if (i > 0) {
                    String op = request.getSearchOp().get(i);
                    if (!op.isEmpty()) {
                        op = op.startsWith("@") ? op.substring(1) : op;
                        // get previous term if present
                        if (!termList.isEmpty()) {
                            String previous = termList.pop();
                            termList.push(previous + " " + op +  " " + term);
                        } else {
                            termList.push(term);
                        }
                    }
                } else {
                    termList.push(term);
                }
            }
        }
        return String.join(" ", termList);
    }

    protected static String trim(String string) {
        return string != null ? string.trim() : "";
    }

    protected static String quote(String string) {
        return '"' +  getEscapedString(string) + '"';
    }

    protected static String getEscapedString(String value) {
        if (value == null) {
            return null;
        }
        return value
                .replace("\\", "\\\\")
                .replace("\n", "\\n")
                .replace("\t", "\\t")
                .replace("\b", "\\b")
                .replace("\r", "\\r")
                .replace("\f", "\\f")
                .replace("\"", "\\\"")
                .replace("'", "\\'");
    }

    public String fromCQLQuery(SearchRequest request) {
        if (request.isValidated()) {
            return request.getSearchValue().getFirst();
        }
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < request.getSearchValue().size(); i++) {
            String key = request.getSearchKey().get(i);
            String query = request.getSearchValue().get(i);
            if (key != null && !key.isEmpty()) {
                if ("all".equals(key)) {
                    query = normalizeValue(query);
                    if (query != null && !query.isEmpty()) {
                        query = phrase(query);
                    }
                } else if (!"cql".equals(key)) {
                    query = normalizeValue(query);
                    if (query != null && !query.isEmpty()) {
                        query = key + " " + getRelation() + " " + phrase(query);
                    }
                }
            }
            if (query != null && !query.isEmpty()) {
                if (!request.getSearchOp().get(i).isEmpty() && !sb.isEmpty()) {
                    sb.append(' ').append(request.getSearchOp().get(i)).append(' ');
                }
                sb.append(query);
            }
        }
        return sb.toString();
    }

    protected String getRelation() {
        return "=";
    }

    protected String phrase(String query) {
        if (query == null) {
            return null;
        }
        // perform auto-phrase if a space is present
        if (query.startsWith("\"") && query.endsWith("\"")) {
            return query;
        }
        if (query.indexOf(" ") > 0) {
            return "\"" + query + "\"";
        }
        return query;
    }

    protected String normalizeValue(String query) {
        return query != null ? query.trim() : null;
    }

    static class TermTransformer implements Visitor {

        private final Map<String, String> keys;

        TermTransformer(Map<String, String> keys) {
            this.keys = keys;
        }

        @Override
        public void visit(SortedQuery sortedQuery) {

        }

        @Override
        public void visit(Query query) {

        }

        @Override
        public void visit(PrefixAssignment prefixAssignment) {

        }

        @Override
        public void visit(ScopedClause scopedClause) {

        }

        @Override
        public void visit(BooleanGroup booleanGroup) {

        }

        @Override
        public void visit(SearchClause searchClause) {

        }

        @Override
        public void visit(Relation relation) {

        }

        @Override
        public void visit(Modifier modifier) {

        }

        @Override
        public void visit(ModifierList modifierList) {

        }

        @Override
        public void visit(Term term) {

        }

        @Override
        public void visit(Identifier identifier) {
        }

        @Override
        public void visit(Index index) {
            String s = index.getContext() + "." + index.getName();
            if (keys.containsKey(s)) {
                String value = keys.get(s);
                if ("null".equals(value) || value == null) {
                    index.setName(null);
                    index.setContext(null);
                } else {
                    int pos = value.lastIndexOf('.');
                    String newContext = pos >= 0 ? value.substring(0, pos) : "";
                    String newName = pos >= 0 ? value.substring(pos + 1) : value;
                    index.setContext(newContext);
                    index.setName(newName);
                }
            }
        }

        @Override
        public void visit(SimpleName simpleName) {

        }

        @Override
        public void visit(SortSpec sortSpec) {

        }

        @Override
        public void visit(SingleSpec singleSpec) {

        }
    }
}
